require "wall";

FLOOR_CONFIG = {};
FLOOR_CONFIG["attic"] = {
	floorLayout = love.filesystem.newFile("asset/floorLayout/attic.csv"),
	floorImage = love.graphics.newImage("asset/image/floor4.png")
};
FLOOR_CONFIG["second-floor"] = {
	floorLayout = love.filesystem.newFile("asset/floorLayout/second-floor.csv"),
	floorImage = love.graphics.newImage("asset/image/floor3.png")
};
FLOOR_CONFIG["main-floor"] = {
	floorLayout = love.filesystem.newFile("asset/floorLayout/main-floor.csv"),
	floorImage = love.graphics.newImage("asset/image/floor2.png")
};
FLOOR_CONFIG["basement"] = {
	floorLayout = love.filesystem.newFile("asset/floorLayout/basement.csv"),
	floorImage = love.graphics.newImage("asset/image/floor1.png")
};

HouseFloor = Class {
	floorImage = nil,
	walls = {}
};

function HouseFloor:loadFloor(floorName)
	self.floorImage = FLOOR_CONFIG[floorName].floorImage;
	
	levelFile = FLOOR_CONFIG[floorName].floorLayout;
	levelFile:open('r');
	fileContents = levelFile:read();
	local lineData = string.explode(fileContents:sub(0, string.len(fileContents) - 2), "\r\n");
	local levelH = #lineData;
	
	for y, line in ipairs(lineData) do
		local data = string.explode(line, ",");
		local levelW = #data;
		for x, value in ipairs(data) do
			local sx = (x - 1) * WALL_WIDTH;
			local sy = (y - 1) * WALL_HEIGHT;
			
			-- Solid Wall
			if(string.find(value, "W") ~= nil) then
				local newWall = Wall(sx, sy);
				table.insert(self.walls, newWall);
			end
		end
	end
end

function HouseFloor:draw()
	local flrImageW, flrImageH = self.floorImage:getWidth(), self.floorImage:getHeight();
	
	love.graphics.setColor(255, 255, 255);
	for x = 0, SCREEN_WIDTH, flrImageW  do
		for y = 0, SCREEN_HEIGHT, flrImageH do
			love.graphics.draw(self.floorImage, x, y);
		end
	end
	
	for index, wall in ipairs(self.walls) do
		wall:draw();
	end
end